! SPT lensing likelihood module, written by S. Feeney based on the
! descriptions provided in arXiv:1202.0546 and at 
! http://pole.uchicago.edu/public/data/vanengelen12/index.html.
! Format based on the ACT lensing likelihood written by J. Dunkley, 
! S. Das, B. Sherwin, R. Hlozek, itself based on the WMAP likelihood 
! code: see Das et al (2011), arXiv:1103.2124

module spt_lensing_likelihood
  
  
  integer, parameter :: l_min_spt      =  50   ! min ell to consider
  integer, parameter :: l_max_spt      =  2050 ! max ell to consider
  integer, parameter :: n_bands        =  14   ! max number of data points
  integer, parameter :: bandwidth      =  100  ! bandwidth
  real(8), parameter :: pi_spt         =  3.141592653589793238462643383279502884197d0
  real(8), parameter :: log_zero       =  -1.0d30
  character(len = *), parameter :: lens_dir = './'
  integer :: min_band, max_band
  integer :: band_l_min(0: n_bands - 1), &
             band_l_max(0: n_bands - 1)
  real(8) :: bpp_dat_spt(0: n_bands - 1), &
             bpp_var_spt(0: n_bands - 1), &
             band_l_mid(0: n_bands - 1), &
             raw_cov_mat(0: n_bands - 1, 0: n_bands - 1), &
             n_l_zero(0: n_bands - 1), &
             fiducial_scaling(0: n_bands - 1), &
             cal_err(0: n_bands - 1, 0: n_bands - 1), &
             fid_ln_det
  logical :: initialise_spt = .true.
  
  private
  
  public :: spt_likelens_init
  public :: spt_likelens_compute
  public :: get_free_lun
  
contains
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  ! read in SPT lensing bandpowers and bins
  subroutine spt_likelens_init
    
    implicit none
    
    integer  :: i, j, lun, status
    real(8), allocatable :: fid_cov_mat(:, :)
    character(len = 255) :: filename, dummy_text
    character(len = 2) :: fmt
    
    
    ! we need to:
    ! 1) read in the SPT lensing bandpowers and error bars
    ! 2) read in the ingredients for building covariance matrices i.e.
    !    i) the raw bandpower covariance,
    !    ii) the zero-order bias and fiducial scaling, and
    !    iii) the calibration-induced uncertainty
    ! 3) check which bins we need to use
    
    
    ! return immediately if routine has already been called
    if (.not. initialise_spt) return
    
    
    ! first read in bandpowers and errors, taking care of the extra
    ! 1 / 2 pi: the bandpowers are quoted as l ^ 4 C_l^{phiphi} / 2 pi
    ! whereas the covariance matrix info is quoted as 
    ! l ^ 4 C_l^{phiphi}
    print*, 'reading SPT lensing bandpowers'
    filename = trim(adjustl(lens_dir)) // 'bandpowers.txt'
    call get_free_lun(lun)
    open(unit = lun, file = trim(adjustl(filename)), &
         form = 'formatted', status = 'old', action = 'read')
    read(lun, *) dummy_text
    do i = 0, n_bands - 1
       
       read(lun, *) band_l_mid(i), bpp_dat_spt(i), bpp_var_spt(i)
       band_l_min(i) = floor(band_l_mid(i) - (bandwidth / 2.d0)) + &
                       1.0d0
       band_l_max(i) = floor(band_l_mid(i) + (bandwidth / 2.d0))
       bpp_dat_spt(i) = 2.d0 * pi_spt * bpp_dat_spt(i)
       bpp_var_spt(i) = (2.d0 * pi_spt * bpp_var_spt(i))**2
       
    enddo
    close(lun)
    
    
    ! read in raw covariance matrix
    print*, 'reading SPT raw covariance matrix'
    filename = trim(adjustl(lens_dir)) // 'covar_l4clphi_1e7.txt'
    call get_free_lun(lun)
    open(unit = lun, file = trim(adjustl(filename)), &
         form = 'formatted', status = 'old', action = 'read')
    write(fmt, '(I2.1)') n_bands
    read(lun, *) dummy_text
    do i = 0, n_bands - 1
       
       read(lun, fmt = '(' // fmt // '(6X, F10.7)' // ')') &
            raw_cov_mat(i, :)
       
    enddo
    close(lun)
    
    
    ! read terms needed to adjust diagonal for sample variance
    print*, 'reading adjustments for sample variance'
    filename = trim(adjustl(lens_dir)) // 'samplevar_scaling.txt'
    call get_free_lun(lun)
    open(unit = lun, file = trim(adjustl(filename)), &
         form = 'formatted', status = 'old', action = 'read')
    read(lun, *) dummy_text
    do i = 0, n_bands - 1
       
       read(lun, *) n_l_zero(i), fiducial_scaling(i)
       
    enddo
    close(lun)
    
    
    ! read in calibration-induced uncertainty
    print*, 'reading SPT calibration-induced uncertainty'
    filename = trim(adjustl(lens_dir)) // 'cal_covar_l4clphi_1e7.txt'
    call get_free_lun(lun)
    open(unit = lun, file = trim(adjustl(filename)), &
         form = 'formatted', status = 'old', action = 'read')
    read(lun, *) dummy_text
    do i = 0, n_bands - 1
       
       read(lun, fmt = '(' // fmt // '(6X, F10.7)' // ')') &
            cal_err(i, :)
       
    enddo
    close(lun)
    
    
    ! form the covariance matrix corresponding to the fiducial LCDM
    ! cosmology and calculate its determinant, which will serve
    ! to normalize the likelihood. the determinant is returned by 
    ! the cov mat inversion subroutine, so use that here (the cov mat
    ! is so small that the overhead in calculating one extra 
    ! inversion is tiny)
    allocate(fid_cov_mat(0: n_bands - 1, 0: n_bands - 1))
    fid_cov_mat(:, :) = raw_cov_mat(:, :) + cal_err(:, :)
    call spt_cov_mat_invert(fid_cov_mat, fid_ln_det, status)
    if (status .ne. 0) then
       
       print*, 'error inverting SPT lensing fiducial covariance ' // &
               'matrix'
       stop
       
    end if
    deallocate(fid_cov_mat)
    
    
    ! define range of bins to be used, given range in ell: this should
    ! be the full range
    do i = 0, n_bands - 1
       
       if(band_l_mid(i) <= l_min_spt) min_band = i + 1
       if(band_l_mid(i) <= l_max_spt) max_band = i
       
    end do
    
    
    ! SPT lensing likelihood is now initialised!
    initialise_spt = .false.
    print*, 'SPT data initialised'
    
    
  end subroutine spt_likelens_init
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! calculate an inverse covariance matrix from a theory C_l^{phiphi} 
  ! (given in bandpowers) and the basic ingredients. also obtain its
  ! determinant
  subroutine spt_cov_mat_compute(bpp_theory, cov_mat, ln_det, status)
    
    implicit none
    
    integer  :: i
    real(8), intent(in) :: bpp_theory(0: n_bands - 1)
    real(8), intent(out) :: cov_mat(0: n_bands - 1, 0: n_bands - 1), &
                            ln_det
    integer, intent(out) :: status
    
    
    ! first re-scale raw covariance diagonal to take account of  
    ! sample variance. as described in the SPT readme, the scaling is
    ! proportional to 
    ! [(L^4C_L^{\phi\phi} + L^2 N_L^(0) ) / 
    !  (L^4 C_L^{\phi\phi}_{fiducial} + L^2N_L^{0})]^2
    ! where "C_L^{\phi\phi}" is the theory bandpower. then add in 
    ! the calibration-induced uncertainty
    cov_mat(:, :) = raw_cov_mat(:, :)
    do i = 0, n_bands - 1
       
       cov_mat(i, i) = cov_mat(i, i) * &
                       ((bpp_theory(i) + n_l_zero(i)) / &
                        fiducial_scaling(i)) ** 2
       
    end do
    cov_mat(:, :) = cov_mat(:, :) + cal_err(:, :)
    
    
    ! invert the covariance matrix using LAPACK
    call spt_cov_mat_invert(cov_mat, ln_det, status)
    
    
  end subroutine spt_cov_mat_compute
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! compute the SPT lensing log-likelihood. the input power spectrum
  ! should be L^4 C_L^{\phi\phi}, with 2 <= L <= L_max
  subroutine spt_likelens_compute(clpp, ln_like)
    
    implicit none
    
    integer :: i, j, status
    real(8), intent(in) :: clpp(2: *)
    real(8), intent(out) :: ln_like
    real(8), allocatable :: cov_mat(:, :)
    real(8) :: bpp_theory(0: n_bands - 1), diffs(0: n_bands - 1), &
               ell(2: band_l_max(n_bands - 1)), &
               work(2: band_l_max(n_bands - 1)), ln_det
    
    
    ! zero log-likelihood to be safe, and calculate ell array
    ln_like = 0.d0
    do i = 2, band_l_max(n_bands - 1)
       
       ell(i) = real(i, kind = 8)
       
    end do
    
    
    ! calculate the ell-weighted average in each bin
    do i = 0, n_bands - 1
       
       bpp_theory(i) = sum(ell(band_l_min(i): band_l_max(i)) * &
                           clpp(band_l_min(i): band_l_max(i))) / &
                       sum(ell(band_l_min(i): band_l_max(i)))
       
    end do
    
    
    ! calculate the covariance matrix and its determinant
    allocate(cov_mat(0: n_bands - 1, 0: n_bands - 1))
    call spt_cov_mat_compute(bpp_theory, cov_mat, ln_det, status)
    
    
    ! if inverse covariance matrix calculation fails (which can only
    ! happen if the inversion fails), assume it is due to an extreme
    ! cosmology and return log_zero
    if (status .ne. 0) then
       
       print*, 'SPT lens warning: rejecting extreme cosmology'
       ln_like = log_zero
       
    else
       
       ! otherwise calculate log-likelihood. subtract off a constant 
       ! offset equal to the log of the determinant of the covariance 
       ! matrix assuming the fiducial LCDM cosmology
       diffs(min_band: max_band) = bpp_dat_spt(min_band: max_band) - &
                                   bpp_theory(min_band: max_band)
       work(:) = 0.d0
       do i = min_band, max_band
          
          work(i) = sum(cov_mat(i, min_band: max_band) * &
                        diffs(min_band: max_band))
          
       end do
       ln_like = sum(diffs(min_band: max_band) * &
                     work(min_band: max_band))
       ln_like = -0.5d0 * (ln_like + ln_det - fid_ln_det)
       
    end if
    deallocate(cov_mat)
    
    
  end subroutine spt_likelens_compute
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! invert a (2D, symmetric, positive-definite) covariance matrix 
  ! using Cholesky decomposition. the LAPACK routine this uses 
  ! sets the upper triangle of the matrix. optionally return the 
  ! (logarithm of the) determinant. supply a status variable if 
  ! you do not want to fail on error
  subroutine spt_cov_mat_invert(two_d, ln_det, status)
    
    implicit none
    
    real(8), dimension(0:, 0:), intent(inout) :: two_d
    real(8), optional :: ln_det
    integer, optional :: status
    real(8), dimension(:), allocatable :: work
    real(8) :: norm, rec_cond
    integer, dimension(:), allocatable :: i_work
    integer :: n_rows, info, i, j
    
    
    ! determine size and norm of matrix
    n_rows = size(two_d, dim = 1)
    call one_norm_2d(two_d, norm)
    
    
    ! calculate Cholesky factorisation of target matrix
    if (present(status)) status = 0
    call dpotrf('U', n_rows, two_d, n_rows, info)
    if (info .ne. 0) then
       
       if (present(status)) then
          
          status = info
          return
          
       else
          
          print*, 'dpotrf failure status ', info
          stop
          
       end if
       
    end if
    
    
    ! optionally calculate the determinant
    if (present(ln_det)) then
       
       call log_determinant(two_d, ln_det)
       
    end if
    
    
    ! estimate condition number
    allocate(work(0: 3 * n_rows))
    allocate(i_work(0: n_rows))
    call dpocon('U', n_rows, two_d, n_rows, norm, rec_cond, &
                work, i_work, info)
    deallocate(work, i_work)
    if (info .ne. 0) then
       
       if (present(status)) then
          
          status = info
          return
          
       else
          
          print*, 'dpocon failure status ', info
          stop
          
       end if
       
    end if
    
    
    ! invert target matrix
    call dpotri('U', n_rows, two_d, n_rows, info)
    if (info .ne. 0) then
       
       if (present(status)) then
          
          status = info
          return
          
       else
          
          print*, 'dpotri failure status ', info
          stop
          
       end if
       
    end if
    
    
    ! fill in missing half
    do i = 0, n_rows - 1
       
       do j = i + 1, n_rows - 1
          
          two_d(j, i) = two_d(i, j)
                  
       end do
       
    end do
    
    
  end subroutine spt_cov_mat_invert
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! calculate the log of the determinant of a matrix which has been 
  ! factorized into A = U^T U. the determinant in this case is the 
  ! product of the squares of the diagonal element: the log-
  ! determinant is therefore the sum of the squares
  subroutine log_determinant(two_d, ln_det)
    
    implicit none
    
    real(8), dimension(0:, 0:), intent(in) :: two_d
    real(8), intent(out) :: ln_det
    integer :: n_rows, i
    
    n_rows = size(two_d, dim = 1)
    ln_det = 0.0d0
    do i = 0, n_rows - 1
       
       ln_det = ln_det + log(two_d(i, i))
       
    end do
    ln_det = 2.0d0 * ln_det
    
  end subroutine log_determinant
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! calculate one-norm of two-dimensional target matrix (this is 
  ! the maximum sum of the absolute values of one column)
  subroutine one_norm_2d(two_d, norm)
    
    implicit none
    
    real(8), dimension(0:, 0:), intent(in) :: two_d
    real(8), intent(out) :: norm
    real(8) :: col_sum
    integer :: n_rows, n_cols, i, j
    
    n_rows = size(two_d, dim = 1)
    n_cols = size(two_d, dim = 1)
    norm = 0.d0
    do j = 0, n_cols - 1
       
       col_sum = 0.d0
       do i = 0, n_rows - 1
          
          col_sum = col_sum + abs(two_d(i, j))
          
       end do
       if (col_sum .gt. norm) norm = col_sum
       
    end do
    
  end subroutine one_norm_2d
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  ! ronseal
  subroutine get_free_lun(lun)
    
    implicit none
    
    integer, intent(out) :: lun
    integer, save :: last_lun = 19
    logical :: used
    
    lun = last_lun
    do
       
       inquire(unit = lun, opened = used)
       if (.not. used) exit
       lun = lun + 1
       
    end do
    last_lun = lun
    
  end subroutine get_free_lun
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
end module spt_lensing_likelihood
